
/* HufBlock.C */

#include <stdio.h>
#include <stdlib.h>

#include "hufblock.h"
#include "types.h"
#include "algo.h"
#include "bitio.h"
#include "bwt.h"
#include "debug.h"
#include "huffman.h"
#include "mtf.h"

/* Constants */

#define BLOCK_SIZE_MIN 256
#define BLOCK_SIZE_MAX 65536
#define BLOCK_SIZE_BIT 16

#define TABLE_NB       256
#define TABLE_NB_MAX   256
#define TABLE_BIT      3

#define WORST_GAIN     -2000000000

#define SYMBOL_NB      257
#define LEN_MAX        25
#define FORMAT         DELTA

/* Types */

typedef struct {
   int Start;
   int End;
   int Size;
   int Table;
   int Freq[SYMBOL_NB];
} block;

typedef struct {
   huftable HufTable[1];
   int Len;
   int Other;
   int Gain;
   int Freq[SYMBOL_NB];
} table;

/* Variables */

ushort *HufBlock;
int     HufBlockSize;

static block Block[TABLE_NB];
static int   BlockNb, BlockSize;

static table Table[TABLE_NB];
static int   TableNb, TableBit;

static uchar TableNo[TABLE_NB];

/* Prototypes */

static void InitBlock   (int Start);
static void UpdateTable (int First, int Second);

/* Functions */

/* AllocHufBlock() */

void AllocHufBlock(void) {

   HufBlock = malloc((size_t)(N*sizeof(int)));
   if (HufBlock == NULL) FatalError("AllocHufBlock(): malloc() = NULL");

   HufBlockSize = 0;
}

/* FreeHufBlock() */

void FreeHufBlock(void) {

   if (HufBlock != NULL) {
      free(HufBlock);
      HufBlock = NULL;
   }
}

/* GetHufBlock() */

void GetHufBlock(void) {

   int I, J, Size;

   HufBlockSize = GetBits(25) + 1;
   BlockSize    = GetBits(BLOCK_SIZE_BIT) + 1;
   BlockNb      = (HufBlockSize + BlockSize - 1) / BlockSize;

   if (Verbosity >= 2) {
      fprintf(stderr,"FullBlockSize = %d\n",HufBlockSize);
      fprintf(stderr,"BlockSize = %d\n",BlockSize);
      fprintf(stderr,"BlockNb = %d\n",BlockNb);
   }

   TableBit = GetBits(TABLE_BIT) + 1;
   TableNb  = GetBits(TableBit) + 1;

   if (Verbosity >= 2) fprintf(stderr,"TableNb = %d\n",TableNb);

   for (I = 0; I < TableNb; I++) {
      AllocHufTable(Table[I].HufTable,SYMBOL_NB,LEN_MAX,FORMAT);
      GetLens(Table[I].HufTable);
   }

   for (I = 0; I < BlockNb; I++) {
      TableNo[I] = 0;
      while (GetBit() == 1) TableNo[I]++;
   }

   DecodeMTF(TableNo,BlockNb);

   Size = 0;

   for (I = 0; I < BlockNb; I++) {
      CompCodes(Table[TableNo[I]].HufTable);
      CompDecodeTable(Table[TableNo[I]].HufTable);
      for (J = 0; J < BlockSize && Size < HufBlockSize; J++) {
	 HufBlock[Size++] = GetHufSym(Table[TableNo[I]].HufTable);
      }
   }

   for (I = 0; I < TableNb; I++) FreeHufTable(Table[I].HufTable);
}

/* SendHufBlock() */

void SendHufBlock(void) {

   int I, J, K, Len, Gain, First, Second, BestLen, BestTable, BestGain;
   int BlockLen;
   int Freq[SYMBOL_NB];
   huftable HufTable[1];

   if (HufBlockSize <= 0) FatalError("HufBlockSize (%d) <= 0 in SendHufBlock()",HufBlockSize);

   BlockSize = (HufBlockSize + TABLE_NB - 1) / TABLE_NB;
   if (BlockSize < BLOCK_SIZE_MIN) BlockSize = BLOCK_SIZE_MIN;
   if (BlockSize > BLOCK_SIZE_MAX) BlockSize = BLOCK_SIZE_MAX;

   BlockNb = 0;
   TableNb = 0;

   for (I = 0; I < HufBlockSize; I += BlockSize) InitBlock(I);

   BlockLen = 0;
   for (I = 0; I < TableNb; I++) BlockLen += Table[I].Len;

   AllocHufTable(HufTable,SYMBOL_NB,LEN_MAX,FORMAT);

   if (Verbosity >= 2) {
      fprintf(stderr,"FullBlockSize = %d\n",HufBlockSize);
      fprintf(stderr,"BlockSize = %d\n",BlockSize);
      fprintf(stderr,"BlockNb = %d\n",BlockNb);
   }

   while (TableNb > 2) {

      BestGain = WORST_GAIN;

      for (I = 0; I < TableNb-1; I++) {
 	 if (Table[I].Other < 0) {
            Table[I].Other = I;
            Table[I].Gain  = WORST_GAIN;
   	    for (J = I+1; J < TableNb; J++) {
	       for (K = 0; K < SYMBOL_NB; K++) {
		  Freq[K] = Table[I].HufTable->HufSym[K].Freq + Table[J].HufTable->HufSym[K].Freq;
	       }
	       CompLens(HufTable,Freq);
	       Gain = Table[I].Len + Table[J].Len - PredictLen(HufTable);
	       if (Gain > Table[I].Gain) {
		  Table[I].Gain  = Gain;
		  Table[I].Other = J;
               }
            }
         }
	 if (Table[I].Gain > BestGain) {
	    BestGain = Table[I].Gain;
	    First    = I;
	    Second   = Table[I].Other;
	 }
      }

      if (TableNb <= TABLE_NB_MAX && BestGain < 0) break;

      BlockLen -= BestGain;

      if (Verbosity >= 3) fprintf(stderr,"TableNb = %4d, BlockSize = %9.2f, TotalLen = %10.2f, Gain = %7.2f\n",TableNb-1,(double)(HufBlockSize)/(double)(TableNb-1),(double)BlockLen/8.0,(double)BestGain/8.0);

      for (K = 0; K < SYMBOL_NB; K++) {
         Freq[K] = Table[First].HufTable->HufSym[K].Freq + Table[Second].HufTable->HufSym[K].Freq;
      }
      CompLens(Table[First].HufTable,Freq);
      Table[First].Len   = PredictLen(Table[First].HufTable);
      Table[First].Other = -1;

      for (I = 0; I < First; I++) {
 	 if (Table[I].Other == First || Table[I].Other == Second) {
 	    Table[I].Other = -1;
         } else {
            if (Table[I].Other > Second) Table[I].Other--;
	    for (K = 0; K < SYMBOL_NB; K++) {
	       Freq[K] = Table[I].HufTable->HufSym[K].Freq + Table[First].HufTable->HufSym[K].Freq;
	    }
	    CompLens(HufTable,Freq);
	    Gain = Table[I].Len + Table[First].Len - PredictLen(HufTable);
	    if (Gain > Table[I].Gain) {
	       Table[I].Gain  = Gain;
	       Table[I].Other = First;
	    }
         }
      }

      for (I = First+1; I < Second; I++) {
 	 if (Table[I].Other == Second) {
            Table[I].Other = -1;
         } else if (Table[I].Other > Second) {
            Table[I].Other--;
         }
      }

      FreeHufTable(Table[Second].HufTable);
      TableNb--;

      for (I = Second; I < TableNb; I++) {
         Table[I] = Table[I+1];
 	 if (Table[I].Other > Second) Table[I].Other--;
      }

      UpdateTable(First,Second);
   }

   if (Verbosity >= 2) fprintf(stderr,"TableNb = %4d, BlockSize = %9.2f, TotalLen = %10.2f\n",TableNb,(double)(HufBlockSize)/(double)(TableNb),(double)BlockLen/8.0);

   FreeHufTable(HufTable);

   /* New optimizing */

   do {

      Gain = FALSE;

      for (I = 0; I < BlockNb; I++) {
	 BestLen   = +2000000000;
	 BestTable = -1;
	 for (J = 0; J < TableNb; J++) {
	    Len = 0;
	    for (K = Block[I].Start; K < Block[I].End; K++) {
	       if (Table[J].HufTable->HufSym[HufBlock[K]].Freq == 0) goto Imp;
	       Len += Table[J].HufTable->HufSym[HufBlock[K]].Len;
	    }
	    if (Len < BestLen) {
	       BestLen   = Len;
	       BestTable = J;
	    }
Imp:        ;
	 }
	 assert(BestTable>=0);
	 if (BestTable != Block[I].Table) {
	    Block[I].Table = BestTable;
	    Gain = TRUE;
	 }
      }

      for (I = 0; I < TableNb; I++) {
	 for (J = 0; J < SYMBOL_NB; J++) Table[I].Freq[J] = 0;
      }

      for (I = 0; I < BlockNb; I++) {
	 for (J = Block[I].Start; J < Block[I].End; J++) {
	    Table[Block[I].Table].Freq[HufBlock[J]]++;
	 }
      }

      for (I = 0; I < TableNb; I++) {
	 CompLens(Table[I].HufTable,Table[I].Freq);
      }

   } while (Gain);

   /* End of new optimizing */

   SendBits(25,HufBlockSize-1);
   SendBits(BLOCK_SIZE_BIT,BlockSize-1);

   for (TableBit = 1; 1<<TableBit < TableNb; TableBit++)
      ;
   SendBits(TABLE_BIT,TableBit-1);
   SendBits(TableBit,TableNb-1);
   for (I = 0; I < TableNb; I++) {
      SendLens(Table[I].HufTable);
      CompCodes(Table[I].HufTable);
   }

   for (I = 0; I < BlockNb; I++) TableNo[I] = Block[I].Table;
   CodeMTF(TableNo,BlockNb);

   for (I = 0; I < BlockNb; I++) {
      for (J = 0; J < TableNo[I]; J++) SendBit(1);
      SendBit(0);
   }

   for (I = 0; I < BlockNb; I++) {
      for (J = Block[I].Start; J < Block[I].End; J++) {
	 SendHufSym(Table[Block[I].Table].HufTable,HufBlock[J]);
      }
   }

   for (I = 0; I < TableNb; I++) {
      CheckFreqs(Table[I].HufTable);
      FreeHufTable(Table[I].HufTable);
   }
}

/* InitBlock() */

static void InitBlock(int Start) {

   int I, End;

   End = Start + BlockSize;
   if (End > HufBlockSize) End = HufBlockSize;

   Block[BlockNb].Start = Start;
   Block[BlockNb].End   = End;
   Block[BlockNb].Size  = End - Start;
   Block[BlockNb].Table = TableNb;

   for (I = 0; I < SYMBOL_NB; I++) Block[BlockNb].Freq[I] = 0;

   for (I = Start; I < End; I++)   Block[BlockNb].Freq[HufBlock[I]]++;
   AllocHufTable(Table[TableNb].HufTable,SYMBOL_NB,LEN_MAX,FORMAT);
   CompLens(Table[TableNb].HufTable,Block[BlockNb].Freq);

   Table[TableNb].Len   = PredictLen(Table[TableNb].HufTable);
   Table[TableNb].Other = -1;

   BlockNb++;
   TableNb++;
}

/* UpdateTable() */

static void UpdateTable(int First, int Second) {

   int I;

   for (I = 0; I < BlockNb; I++) {
      if (Block[I].Table == Second) {
         Block[I].Table = First;
      } else if (Block[I].Table > Second) {
         Block[I].Table--;
      }
   }
}

/* End of HufBlock.C */

